#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

from math import pi

epsilon_0 = 8.8541878128e-12
mu_0 = 1.25663706212e-06
c = 299792458.0
